// gcc -Wall -Wpedantic -Wextra -std=c17 -march=haswell -O3 random_access.c

#define _POSIX_C_SOURCE 199309L
#include <time.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#define INDEX_T uint32_t
#define DATA_T int64_t


void shuffle_array(INDEX_T* indices, INDEX_T length) {
    printf("Shuffling indices...\n");
    for (INDEX_T i = 0; i < length; i++) {
        // pick a random element to the right of element i, and swap
        INDEX_T p = i + (rand() % (length - i));
        DATA_T tmp = indices[p];
        indices[p] = indices[i];
        indices[i] = tmp;
    }
}


void gently_shuffle_array(INDEX_T* indices, INDEX_T length) {
    printf("Shuffling indices, but just a little...\n");
    INDEX_T limit = 10;
    // shuffle array indicies with nearby values
    for (INDEX_T i = 0; i < length - limit - 1; i++) {
        INDEX_T offset = rand() % limit;
        DATA_T tmp = indices[i + offset];
        indices[i + offset] = indices[i];
        indices[i] = tmp;
    }
}


void two_sweep_array(INDEX_T* indices, INDEX_T length) {
    printf("Replacing indices with 'sweep with two cursors'...\n");
    for (INDEX_T i = 0; i < length - 1; i += 2) {
        indices[i] = i / 2;
        indices[i + 1] = length / 2 + i / 2;
    }
}


void build_arrays(DATA_T** data, INDEX_T** indices, INDEX_T length) {
    printf("Allocating %g + %g MB...\n", (double)length * sizeof(DATA_T) / 1024 / 1024, (double)length * sizeof(INDEX_T) / 1024 / 1024);
    *data = (DATA_T*)malloc(length * sizeof(DATA_T));
    *indices = (INDEX_T*)malloc(length * sizeof(INDEX_T));
    if (data == NULL || indices == NULL) {
        printf("Couldn't allocate.\n");
        return;
    }

    printf("Filling arrays...\n");
    for (INDEX_T i = 0; i < length; i++) {
        (*indices)[i] = i;
        (*data)[i] = rand() % (1 << 10);
    }
}


void sum_array(DATA_T* data, INDEX_T* indices, INDEX_T length) {
    printf("Calculating sum...\n");
    struct timespec start, end;
    DATA_T total = 0;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    for (INDEX_T i = 0; i < length; i++) {
        total += data[indices[i]];
    }
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) * 1e9 + (end.tv_nsec - start.tv_nsec);
    printf("Calculated sum of %u values %lu in %.0f ns, %.2f ns/elt.\n", length, total, elapsed, elapsed / length);
}


int test(INDEX_T length) {
    srand(time(NULL));
    DATA_T* data;
    INDEX_T* indices;
    build_arrays(&data, &indices, length);
    if (data == NULL || indices == NULL) {
        return 1;
    }

#ifndef NO_SHUFFLE
    shuffle_array(indices, length);
#endif
#ifdef GENTLY_SHUFFLE
    gently_shuffle_array(indices, length);
#endif
    //two_sweep_array(indices, length);

    sum_array(data, indices, length);

    free(data);
    free(indices);
    return 0;
}


int main(int argc, char* argv[]) {
    INDEX_T length;
    if (argc < 3) {
        length = 1 << 24;
    } else {
        INDEX_T base = atol(argv[1]);
        INDEX_T shift = atol(argv[2]);
        length = base << shift;
    }
    return test(length);
}
